@@ -1,9 +1,14 @@
--- created by Martin Hawlisch
+--
+-- A level for Enigma
+--
+-- Copyright (c) 2003 Martin Hawlisch
+-- Licensed under the GPL version 2.
+
 
 create_world( 20, 13)
 oxyd_default_flavor = "b"
 
-fill_floor("fl-rough", 0,0, level_width,level_height)
+fill_floor("fl-rough")
 fill_floor("fl-rough_slow", 14, 5, 3, 3)
 
 stone = "st-rock2"
@@ -42,6 +47,5 @@ oxyd(12, 6)
 
 oxyd_shuffle()
 
-set_actor("ac-blackball", 9.2, 6.5, {player=0})
-set_actor("ac-whiteball-small", 15.5, 6.5, {player=0, mouseforce=1})
-
+set_actor("ac-blackball", 9.2, 6.5, {essential=1})
+set_actor("ac-whiteball-small", 15.5, 6.5, {player=0})
