@@ -621,12 +621,14 @@ namespace
 
         display::SpriteHandle m_halosprite;
         double                m_shield_rest_time;
-        static const double   SHIELD_TIME = 10.0;
+        static const double   SHIELD_TIME;
         HaloState             m_halostate;
 
         double m_drunk_rest_time;
         double m_invisible_rest_time;
     };
+
+    const double BasicBall::SHIELD_TIME = 5.0;
 }
 
 BasicBall::BasicBall(const ActorTraits &tr)
