@@ -25,6 +25,8 @@
 
 #include "stones_internal.hh"
 
+#include "px/tools.hh"
+
 #include <algorithm>
 #include <cassert>
 #include <iostream>
@@ -44,8 +46,8 @@ namespace
         {}
 
     private:
-        static const double RATE          = 1.0;
-        static const double IMPULSE_DELAY = 0.1;
+        static const double RATE;
+        static const double IMPULSE_DELAY;
 
         bool clockwise;
         bool movable;
@@ -100,6 +102,9 @@ namespace
             init_model();
         }
     };
+
+    const double RotatorStone::RATE          = 1.0;
+    const double RotatorStone::IMPULSE_DELAY = 0.1;
 }
 
 
@@ -763,15 +768,15 @@ namespace {
     class ConnectiveStone : public Stone {
     public:
         ConnectiveStone(const char *kind, int connections)
-            : Stone(kind)
+        : Stone(kind)
         {
             set_attrib("connections", connections);
         }
 
         DirectionBits get_connections() const;
-
-    private:
+    protected:
         void init_model();
+    private:
         virtual int get_modelno() const;
     };
 }
@@ -805,7 +810,7 @@ namespace
         CLONEOBJ(BigBrick);
     public:
         BigBrick(int connections)
-            : ConnectiveStone("st-bigbrick", connections)
+        : ConnectiveStone("st-bigbrick", connections)
         {}
     };
 }
@@ -815,10 +820,10 @@ namespace
 
 /** \page st-puzzle Puzzle Stone
 
-These stones can be connected to other stones of the same type.  Any
-of the four faces of the stone can have ``socket''.  If the touching
-faces of two adjacent stones both have a socket, the two stones link
-up and henceforth move as group.
+Puzzle stones can be connected to other stones of the same type.  Any
+of the four faces of the stone can have ``socket''.  If the adjoining
+faces of two neighboring stones both have a socket, the two stones
+link up and henceforth move as group.
 
 A cluster of puzzle stones may for example look like this:
 
@@ -839,9 +844,7 @@ cluster.  A cluster is complete if none of its stones has an
 unconnected socket.
 
 When touched with a magic wand the puzzle stones rotate
-row- or columnwise. If one of these rows/columns contains exactly
-one stone, then that stone is moved (regardless whether it is a
-member of a cluster or not).
+row- or columnwise.  
 
 \subsection puzzlea Attributes
 
@@ -918,8 +921,6 @@ namespace
         void        explode();
         bool        explode_complete_cluster();
 
-        DirectionBits calc_illumination();
-
         /* ---------- TimeHandler interface ---------- */
 
         void alarm();
@@ -936,6 +937,7 @@ namespace
         void on_creation (GridPos p);
         void on_removal (GridPos p);
         void on_impulse (const Impulse& impulse);
+        void on_laserhit (Direction dir);
 
         bool is_floating() const;
 
@@ -1177,7 +1179,6 @@ bool PuzzleStone::cluster_complete() {
     return find_cluster(c);
 }
 
-
 int PuzzleStone::get_modelno() const {
     int modelno = int_attrib("connections");
     if (oxyd1_compatible()) modelno += 16;
@@ -1187,20 +1188,18 @@ int PuzzleStone::get_modelno() const {
 void PuzzleStone::rotate_cluster(const Cluster &c) {
     unsigned size = c.size();
     if (size > 1) {
-        vector<Stone*> stones;
-        for (unsigned i=0; i<size; ++i)
-            stones.push_back(YieldStone(c[i]));
-
-        // exclude last list item
-        for (unsigned i=0; i<size-1; ++i)
-            SetStone(c[i+1], stones[i]);
-        SetStone(c[0], stones.back());
+        int cn = GetStone(c[size-1])->int_attrib("connections");
+        for (unsigned i=size-1; i>0; --i) {
+            PuzzleStone *st = dynamic_cast<PuzzleStone*> (GetStone (c[i]));
+            st->set_attrib ("connections", GetStone(c[i-1])->int_attrib ("connections"));
+            st->init_model();
+        }
+        GetStone(c[0])->set_attrib ("connections", cn);
+        dynamic_cast<PuzzleStone*> (GetStone(c[0]))->init_model();
     }
 }
 
-StoneResponse
-PuzzleStone::collision_response(const StoneContact &/*sc*/)
-{
+StoneResponse PuzzleStone::collision_response(const StoneContact &/*sc*/) {
     if (get_connections() == NODIRBIT)
         return STONE_PASS;
     return STONE_REBOUND;
@@ -1224,16 +1223,12 @@ void PuzzleStone::trigger_explosion_at (GridPos p, double delay,
 }
 
 void PuzzleStone::explode() {
-    GridPos       p     = get_pos();
-//     DirectionBits dir   = get_connections();
-//     const char    *expl = (dir == NODIRBIT || dir == ALL_DIRECTIONS)
-//         ? "it-explosion2" : "it-explosion1";
-    int          ox_attr       = int_attrib("oxyd");
+    GridPos p       = get_pos();
+    int     ox_attr = int_attrib("oxyd");
 
     // exchange puzzle stone with explosion
     sound_event("stonedestroy");
-//    KillStone(p);
-    SetStone(p, MakeStone("st-explosion")); //expl));
+    SetStone(p, MakeStone("st-explosion"));
 
     // trigger all adjacent puzzle stones :
     const double DEFAULT_DELAY = 0.2;
@@ -1291,9 +1286,8 @@ void PuzzleStone::on_impulse(const Impulse& impulse)
         bool    is_complete     = find_cluster(c);
         bool    actor_with_wand = false;
 
-        if (Actor *ac = dynamic_cast<Actor*>(impulse.sender)) {
+        if (Actor *ac = dynamic_cast<Actor*>(impulse.sender)) 
             actor_with_wand = player::wielded_item_is(ac, "it-magicwand");
-        }
 
         maybe_move_cluster(c, is_complete, actor_with_wand, impulse.dir);
     }
@@ -1380,56 +1374,36 @@ void PuzzleStone::maybe_rotate_cluster(Direction dir)
     }
 }
 
-void PuzzleStone::on_creation (GridPos p) 
-{
+void PuzzleStone::on_creation (GridPos p) {
     photo_activate();
     ConnectiveStone::on_creation (p);
-    illumination = DirectionBits(ALL_DIRECTIONS+1); // invalid
+    illumination = NODIRBIT;
 }
 
-void PuzzleStone::on_removal(GridPos p) 
-{
+void PuzzleStone::on_removal(GridPos p) {
     photo_deactivate();
     ConnectiveStone::on_removal(p);
 }
 
+void PuzzleStone::on_laserhit (Direction dir) {
+    px::set_flags (illumination, to_bits(reverse(dir)));
+}
+
 void PuzzleStone::on_recalc_start() {
+    illumination = NODIRBIT;
 }
 
-DirectionBits
-PuzzleStone::calc_illumination()
-{
-    DirectionBits illu = NODIRBIT;
-    GridPos       p    = get_pos();
-
-    if (lasers::LightFrom(p, NORTH)) illu = DirectionBits(illu|NORTHBIT);
-    if (lasers::LightFrom(p, SOUTH)) illu = DirectionBits(illu|SOUTHBIT);
-    if (lasers::LightFrom(p, EAST))  illu = DirectionBits(illu|EASTBIT);
-    if (lasers::LightFrom(p, WEST))  illu = DirectionBits(illu|WESTBIT);
-
-    return illu;
-}
-
-void PuzzleStone::on_recalc_finish() 
-{
-    DirectionBits illu = calc_illumination();
-
-    if (illu != illumination) {
-        if (illumination != (ALL_DIRECTIONS+1)) { // not a new stone
-            DirectionBits newBeams = DirectionBits(illu ^ (illu & illumination));
-
-            if (state == IDLE && newBeams != NODIRBIT) {
-                if (!explode_complete_cluster()) {
-                    if (oxyd1_compatible()) {
-                        if (newBeams & NORTHBIT) maybe_rotate_cluster(SOUTH);
-                        if (newBeams & SOUTHBIT) maybe_rotate_cluster(NORTH);
-                        if (newBeams & EASTBIT)  maybe_rotate_cluster(WEST);
-                        if (newBeams & WESTBIT)  maybe_rotate_cluster(EAST);
-                    }
-                }
-            }
+void PuzzleStone::on_recalc_finish() {
+    if (illumination != (ALL_DIRECTIONS+1) && 
+        illumination != NODIRBIT && 
+        state == IDLE) 
+    {
+        if (!explode_complete_cluster() && oxyd1_compatible()) {
+            if (illumination & NORTHBIT) maybe_rotate_cluster(SOUTH);
+            if (illumination & SOUTHBIT) maybe_rotate_cluster(NORTH);
+            if (illumination & EASTBIT)  maybe_rotate_cluster(WEST);
+            if (illumination & WESTBIT)  maybe_rotate_cluster(EAST);
         }
-        illumination = illu;
     }
 }
 
